/*******************************************************************************
    Copyright (c) 2019 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "nv-linux.h"
#include "nv-caps.h"
#include "nv-procfs.h"

/* sys_close() or __close_fd() */
#include <linux/syscalls.h>

struct nv_cap
{
    char *name;
    struct proc_dir_entry *parent;
    struct proc_dir_entry *entry;
};

typedef struct nv_cap_file_private
{
    /* Will be used for fork() detection.*/
    struct files_struct *files;
} nv_cap_file_private_t;

static int nv_cap_procfs_open(struct inode *inode, struct file *file)
{
    nv_cap_file_private_t *private = NULL;

    NV_KMALLOC(private, sizeof(nv_cap_file_private_t));
    if (private == NULL)
    {
        return -ENOMEM;
    }

    /* Just copy over data for "fd" validation */
    private->files = current->files;
    file->private_data = private;

    return 0;
}

static int nv_cap_procfs_release(struct inode *inode, struct file *file)
{
    if (file->private_data != NULL)
    {
        NV_KFREE(file->private_data, sizeof(nv_cap_file_private_t));
    }

    /*
     * All open files using the proc entry will be invalidated
     * if the entry is removed.
     */
    file->private_data = NULL;
    return 0;
}

static nv_proc_ops_t g_nv_cap_procfs_fops = {
    NV_PROC_OPS_SET_OWNER()
    .NV_PROC_OPS_OPEN    = nv_cap_procfs_open,
    .NV_PROC_OPS_RELEASE = nv_cap_procfs_release,
};

int NV_API_CALL nv_cap_validate_and_dup_fd(const nv_cap_t *cap, int fd)
{
    struct file *file;
    int dup_fd;
    const nv_cap_file_private_t *private = NULL;
    struct inode *inode = NULL;

    if (cap == NULL)
    {
        return -1;
    }

    file = fget(fd);
    if (file == NULL)
    {
        return -1;
    }

    inode = NV_FILE_INODE(file);

    /*
     * Make sure the fd belongs to procfs and validate the associated
     * capability; the NV_PDE_DATA(inode) check ensures this file was created
     * by nv_cap_create_file_entry().
     */
    if ((inode == NULL) ||
        (strcmp(inode->i_sb->s_type->name, "proc") != 0) ||
        (NV_PDE_DATA(inode) != cap))
    {
        goto err;
    }

    private = (nv_cap_file_private_t *)file->private_data;

    /* Make sure the fd is not duped during fork() */
    if ((private == NULL) ||
        (private->files != current->files))
    {
        goto err;
    }

    dup_fd = NV_GET_UNUSED_FD_FLAGS(O_CLOEXEC);
    if (dup_fd < 0)
    {
          dup_fd = NV_GET_UNUSED_FD();
          if (dup_fd < 0)
          {
               goto err;
          }
    }

    fd_install(dup_fd, file);
    return dup_fd;

err:
    fput(file);
    return -1;
}

void NV_API_CALL nv_cap_close_fd(int fd)
{
    if (fd == -1)
    {
        return;
    }

    /*
     * Acquire task_lock as we access current->files explicitly (__close_fd)
     * and implicitly (sys_close), and it will race with the exit path.
     */
    task_lock(current);

    /* Nothing to do, we are in exit path */
    if (current->files == NULL)
    {
        task_unlock(current);
        return;
    }

/*
 * From v4.17-rc1 kernels have stopped exporting sys_close(fd) and started
 * exporting __close_fd, as of this commit:
 * 2018-04-02 2ca2a09d6215 ("fs: add ksys_close() wrapper; remove in-kernel
 *  calls to sys_close()")
 */
#if NV_IS_EXPORT_SYMBOL_PRESENT___close_fd
    __close_fd(current->files, fd);
#else
    sys_close(fd);
#endif

    task_unlock(current);
}

nv_cap_t* NV_API_CALL nv_cap_create_file_entry(nv_cap_t *parent_cap,
                                               const char *name, int mode)
{
    nv_cap_t *cap = NULL;

    if (parent_cap == NULL || name == NULL)
    {
        return NULL;
    }

    NV_KMALLOC(cap, sizeof(nv_cap_t));
    if (cap == NULL)
    {
        return NULL;
    }

    NV_KMALLOC(cap->name, strlen(name) + 1);
    if (cap->name == NULL)
    {
        NV_KFREE(cap, sizeof(nv_cap_t));
        return NULL;
    }

    sprintf(cap->name, name);
    cap->parent = parent_cap->entry;
    mode |= S_IFREG;

    /* Set (void*)cap as proc data. Will be used for "fd" validation. */
    cap->entry = proc_create_data(name, mode, parent_cap->entry,
                                   &g_nv_cap_procfs_fops, (void*)cap);
    if (cap->entry == NULL)
    {
        NV_KFREE(cap->name, strlen(cap->name) + 1);
        NV_KFREE(cap, sizeof(nv_cap_t));
        return NULL;
    }

    return cap;
}

nv_cap_t* NV_API_CALL nv_cap_create_dir_entry(nv_cap_t *parent_cap,
                                              const char *name, int mode)
{
    nv_cap_t *cap = NULL;

    if (parent_cap == NULL || name == NULL)
    {
        return NULL;
    }

    NV_KMALLOC(cap, sizeof(nv_cap_t));
    if (cap == NULL)
    {
        return NULL;
    }

    NV_KMALLOC(cap->name, strlen(name) + 1);
    if (cap->name == NULL)
    {
        NV_KFREE(cap, sizeof(nv_cap_t));
        return NULL;
    }

    sprintf(cap->name, name);
    cap->parent = parent_cap->entry;
    mode |= S_IFDIR;

    cap->entry = NV_PROC_MKDIR_MODE(name, mode, parent_cap->entry);
    if (cap->entry == NULL)
    {
        NV_KFREE(cap->name, strlen(cap->name) + 1);
        NV_KFREE(cap, sizeof(nv_cap_t));
        return NULL;
    }

    return cap;
}

nv_cap_t* NV_API_CALL nv_cap_init(const char *path)
{
    nv_cap_t parent_cap;
    nv_cap_t *cap;
    int mode;
    char *name = NULL;
    char dir[] = "/capabilities";

    if (path == NULL)
    {
        return NULL;
    }

    NV_KMALLOC(name, (strlen(path) + strlen(dir)) + 1);
    if (name == NULL)
    {
        return NULL;
    }

    strcpy(name, path);
    strcat(name, dir);
    parent_cap.entry = NULL;
    mode =  S_IRUGO | S_IXUGO;
    cap = nv_cap_create_dir_entry(&parent_cap, name, mode);

    NV_KFREE(name, strlen(name) + 1);
    return cap;
}

void NV_API_CALL nv_cap_destroy_entry(nv_cap_t *cap)
{
    if (WARN_ON(cap == NULL))
    {
        return;
    }

    remove_proc_entry(cap->name, cap->parent);
    NV_KFREE(cap->name, strlen(cap->name) + 1);
    NV_KFREE(cap, sizeof(nv_cap_t));
}
